# built-in
from datetime import datetime
import os

# third-party
import numpy as np

# local
from epibox import config_debug


def write_acq_file(a_file, t, fmt):
    np.savetxt(
        a_file,
        t,
        fmt=fmt,
        delimiter="	",
        newline="\n",
        header="",
        footer="",
        comments="",
    )


def write_annot_file(recording_name, annot):

    annot_filename = os.path.join(os.path.split(recording_name)[
        0], "annotations" + ".txt")

    if not os.path.isfile(annot_filename):
        header = 'filename\tlabel\tvalue\treal time\tEpiBOX time\n'
    else:
        header = ''

    with open(annot_filename, "a+") as file:
        file.write(
            f"{header}{os.path.split(recording_name)[1]}\t{annot[0]}\t{annot[1]}\t{annot[2]}\t{annot[3]}\n"
        )


def write_summary_file(recording_name):

    duration = datetime.now() - datetime.strptime(
        os.path.split(recording_name)[1][1:-4], "%Y-%m-%d %H-%M-%S"
    )
    config_debug.log(f"duration: {str(duration)}")

    with open(
        os.path.join(os.path.split(recording_name)
                     [0], "summary" + ".txt"), "a+"
    ) as file:
        file.write(
            "{}  {}\n".format(
                os.path.split(recording_name)[1], str(duration).split(".")[0]
            )
        )



# built-in
import time
import sys
from datetime import datetime

# third-party
import numpy as np

# local
from epibox.common.read_modules import read_modules
from epibox.common.write_file import write_acq_file


def run_system(
    devices,
    a_file,
    sync_param,
    mac_channels,
    sensors,
    save_fmt,
    header,
    client,
):

    # Read batch of samples from the devices and save to the active session's file  ===============================

    if time.time() - sync_param["strtime"] > 5:

        sync_param["strtime"] = time.time()

    for i, device in enumerate(devices):
        sync_param["sync_arr_" + chr(ord("@") + i + 1)] = np.zeros(1000, dtype=float)

    now = datetime.now()
    sync_param["sync_time"] = now.strftime("%Y-%m-%d %H:%M:%S.%f")

    # read batch of samples from devices
    t, t_str, t_display = read_modules(devices, mac_channels, sensors, header)

    i = time.time() - sync_param["inittime"]

    # config_debug.log elapsed time
    sys.stdout.write("\rElapsed time (seconds): % i " % i)
    sys.stdout.flush()

    # Write batch of samples to file
    write_acq_file(a_file, t, save_fmt)

    return t, t_display, a_file, sync_param



import json
import ast
import os
import time
from pathlib import Path


def get_default():

    defaults_dir = os.path.join(
        Path.home(), "Documents", "EpiBOX Core")
    defaults_path = os.path.join(defaults_dir, "args.json")

    defaults = {
        "initial_dir": "EpiBOX Core",
        "fs": 1000,
        "channels": [],
        "devices_mac": {"MAC1": "34:94:54:5E:32:2E"},
        "save_raw": "true",
        "patient_id": "default",
        "service": "scientisst",
    }

    if os.path.isdir(defaults_dir):

        if os.path.isfile(defaults_path):
            with open(defaults_path, "r") as json_file:
                defaults = json_file.read()
                defaults = ast.literal_eval(defaults)

    else:  # the first time using EpiBOX Core, there will be no default file
        os.makedirs(defaults_dir)

    with open(defaults_path, "w+") as json_file:
        json.dump(defaults, json_file)

    return defaults


def set_new_default_item(item_key, item):

    defaults = get_default()
    defaults[item_key] = item

    with open(
        os.path.join(Path.home(), "Documents",
                     "EpiBOX Core", "args.json"), "w+"
    ) as json_file:
        json.dump(defaults, json_file)

    time.sleep(5)



# third-party
from scipy import signal
import itertools
import numpy as np

# local
from epibox import config_debug


def get_factors(n):

    n2 = n
    i = 2
    factors = []
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n = i
            factors.append(i)
    if n > 1:
        factors.append(n)

    factors.sort(reverse=True)
    final_factors = [f for f in factors if f > 13]
    to_combine = [f for f in factors if f < 13]
    combos = list(itertools.combinations(to_combine, 2))

    while np.prod(final_factors) != n2:

        # config_debug.log(f"combos: {combos}")
        combination = [
            [i, p] for i, p in enumerate([x * y for x, y in combos]) if p <= 13
        ]

        final_factors += [combination[0][1]]
        # config_debug.log(f"final_factors: {final_factors}")

        combos.remove(combos[combination[0][0]])

    return final_factors


def decimate(t, fs):
    # if phisiological signal downsample to 100Hz, if acc decimate 10Hz
    t_display = []

    for i in range(t.shape[1]):
        t_aux = t[:, i]
        if fs >= 1000:
            t_aux = signal.decimate(t_aux, 10)

        t_display += [t_aux.tolist()]

    return t_display



class Flags:
    """
    Class debug
    """

    def __init__(self):

        self.debug = True


def log(message):
    """
    config_debug.logs a message only if debug is true
    """
    flag = Flags()
    if flag.debug:
        print(message)



# third-party
import bitalino
import scientisst
from serial.tools.list_ports import comports

# local
from epibox import config_debug
from epibox.exceptions.system_exceptions import (
    BITalinoParameterError,
    MQTTConnectionError,
    ScientISSTConnectionError,
    ScientISSTNotFound,
)


def connect_device(address, client, devices, service):

    connected = False
    devices = [d for d in devices if d]  # remove None

    if address in [d.address for d in devices]:
        try:
            config_debug.log(
                f"{address} state: {[d.state()for d in devices if d.address==address]}"
            )
            connected = True

        except AttributeError:
            del devices[[d.address for d in devices].index(address)]

    else:

        if service == "bitalino":
            device = bitalino.BITalino(address, timeout=5)
            setattr(device, "address", address) # Bitalino uses "address" attribute and ScientISST uses "address"
            devices += [device]

        elif service == "scientisst":

            # TODO: implement for windows - find COM port through MAC
            # ports = comports()
            # address = [port.device for port in ports if address in port.device]

            # if address == []:
            #     raise ScientISSTNotFound
            # else:
            #     address = address[0]

            device = scientisst.ScientISST(address)
            devices += [device]

    devices = [d for d in devices if d]  # remove None

    if not connected or address not in [d.address for d in devices]:
        message_info = client.publish(
            topic="rpi",
            qos=2,
            payload="['MAC STATE', '{}', '{}']".format(address, "failed"),
        )
        if message_info.rc == 4:
            raise MQTTConnectionError
    else:
        message_info = client.publish(
            topic="rpi",
            qos=2,
            payload="['MAC STATE', '{}', '{}']".format(address, "connected"),
        )
        if message_info.rc == 4:
            raise MQTTConnectionError

    return connected, devices



# built-in
import json

# local
from epibox.common.get_defaults import set_new_default_item
from epibox.devices.manage_devices import connect_devices, start_devices
from epibox.exceptions.exception_manager import (
    handle_case_6,
    kill_case_1,
    kill_case_2,
    kill_case_3,
    kill_case_4,
    kill_case_5,
    kill_case_7,
)
from epibox.common.setup import setup_client, setup_config, setup_variables
from epibox.common.create_folder import create_folder
from epibox.common.open_file import open_file
from epibox.common.write_file import write_annot_file
from epibox.common.run_system import run_system
from epibox.common import process_data
from epibox import config_debug
from epibox.exceptions.system_exceptions import (
    BITalinoParameterError,
    DeviceConnectionTimeout,
    DeviceNotIDLEError,
    DeviceNotInAcquisitionError,
    MQTTConnectionError,
    PlatformNotSupportedError,
    StorageTimeout,
)
from epibox.mqtt_manager.message_handler import send_default


# ****************************** MAIN SCRIPT ***********************************
# This is the main script. It is lauched by EpiBOX Core, through the epibox_startup.sh Bash script.

# This script attempts to connect to the default biosignal acquisition devices in a continuous loop.
# The loop stops only if:
#       - connection is successful
#       - timeout is achieved (2min)

# If timeout, the script ends (being later lauched again by EpiBOX Core)
# If connection is successful, acquisition starts and PyEpiBOX receives and handles data from the
# acquisition devices.
# The loop stops only if:
#       - user stops the acquisition
#       - at least one of the devices disconnects


def main():

    devices = []
    # Setup MQTT client | read default configurations | initiate variables ===========================
    try:
        client = setup_client()

    except (MQTTConnectionError, ConnectionRefusedError, ValueError, TypeError) as e:
        config_debug.log(e)
        kill_case_1()
    except (TimeoutError, ConnectionAbortedError):
        config_debug.log("Wrong network - connect to PreEpiSeizures")
        kill_case_1()

    try:
        opt, channels, sensors, service, save_raw = setup_config(client)

        (
            t_all,
            already_notified_pause,
            system_started,
        ) = setup_variables()  # raises no errors

        # Create folder with patient ID
        directory = create_folder(
            opt["initial_dir"], "{}".format(opt["patient_id"]))

    except MQTTConnectionError as e:
        kill_case_1()
    except PermissionError:
        client.publish("rpi", str(["STORAGE ERROR"]))
        set_new_default_item(item_key="initial_dir", item="EpiBOX Core")
        send_default(client)
        kill_case_2(client)
    except (
        ConnectionRefusedError,
        ValueError,
        TypeError,
        OSError,
        StorageTimeout,
        PlatformNotSupportedError,
    ) as e:
        config_debug.log(e)
        kill_case_2(client)

    already_timed_out = False

    # Start loop to connect PyEpiBOX to acquisition devices =========================================
    if not client.keepAlive:
        kill_case_2()
    try:
        devices = connect_devices(  # exceptions handled inside
            client, devices, opt, already_timed_out,
        )

    except (
        DeviceConnectionTimeout,
        ValueError,
        TypeError,
        BITalinoParameterError,
    ) as e:
        config_debug.log(e)
        kill_case_3(client, devices)

    except MQTTConnectionError as e:
        config_debug.log(e)
        kill_case_4(devices)

    config_debug.log("Initial setup complete!")

    if not client.keepAlive:
        kill_case_4(devices)

    while client.keepAlive:

        try:

            if client.pauseAcq and not already_notified_pause:
                # Pause acquisition if command is received via MQTT ========================================

                # devices = pause_devices(client, devices)
                system_started = handle_case_6(devices, a_file, system_started)
                message_info = client.publish("rpi", str(["PAUSED"]))
                if message_info.rc == 4:
                    raise MQTTConnectionError
                already_notified_pause = True

            if not client.pauseAcq:

                if already_notified_pause:
                    message_info = client.publish("rpi", str(["RECONNECTING"]))
                    if message_info.rc == 4:
                        raise MQTTConnectionError
                    already_notified_pause = False

                if not system_started:
                    if not client.keepAlive:
                        kill_case_4()

                    # If the devices have not yet started acquiring or they are paused, start acquisition
                    a_file, save_fmt, header = open_file(
                        directory,
                        devices,
                        channels,
                        sensors,
                        opt["fs"],
                        save_raw,
                        service,
                    )

                    if client.newAnnot != None:
                        # Write user annotation to file if one is received via MQTT ===============================
                        write_annot_file(a_file.name, client.newAnnot)
                        client.newAnnot = None

                    sync_param = start_devices(
                        client, devices, opt["fs"], channels, header
                    )
                    system_started = True
                    already_timed_out = False

                if client.newAnnot != None:
                    # Write user annotation to file if one is received via MQTT ===============================
                    write_annot_file(a_file.name, client.newAnnot)
                    client.newAnnot = None

                if not client.keepAlive:
                    kill_case_5(client, devices, a_file)

                # Read batch of samples from the acquisition devices and store on the active session's file
                _, t_disp, a_file, sync_param = run_system(
                    devices,
                    a_file,
                    sync_param,
                    channels,
                    sensors,
                    save_fmt,
                    header,
                    client,
                )

                # Subsample batch of samples and send to the EpiBOX App for visualization purposes ================
                t_display = process_data.decimate(t_disp, opt["fs"])
                t_all += t_display[0]
                json_data = json.dumps(["DATA", t_display, channels, sensors])
                message_info = client.publish("rpi", json_data)
                if message_info.rc == 4:
                    raise MQTTConnectionError

                already_timed_out = False

        except (
            DeviceNotIDLEError,
            BITalinoParameterError,
            FileNotFoundError,
            PermissionError,
            OSError,
            KeyboardInterrupt,
        ) as e:
            config_debug.log(e)
            kill_case_5(client, devices, a_file)

        except DeviceNotInAcquisitionError as e:
            config_debug.log(e)
            message_info = client.publish("rpi", str(["RECONNECTING"]))
            if message_info.rc == 4:
                raise MQTTConnectionError
            system_started = handle_case_6(devices, a_file, system_started)
            continue

        except MQTTConnectionError as e:
            config_debug.log(e)
            kill_case_7(client, devices, a_file)

    # =========================================================================================================



# built-in
import time
import os
from sys import platform
from pathlib import Path

# third-party
import paho.mqtt.client as mqtt
from epibox.common.get_defaults import get_default
from epibox.exceptions.system_exceptions import (
    MQTTConnectionError,
    PlatformNotSupportedError,
    StorageTimeout,
)

# local
from epibox.mqtt_manager.message_handler import on_message, send_default
from epibox.mqtt_manager.utils import random_str
from epibox import config_debug


def setup_client():
    # Set up MQTT client =========================================================================

    client_name = random_str(6)
    config_debug.log(f"Client name (acquisition): {client_name}")
    host_name = "192.168.0.10"
    topic = "rpi"

    # raises ValueError and ConnectionRefusedError
    client = mqtt.Client(client_name)

    setattr(client, "keepAlive", True)
    setattr(client, "pauseAcq", False)
    setattr(client, "newAnnot", None)

    client.username_pw_set(username="preepiseizures",
                           password="preepiseizures")
    client.connect(host_name)  # raises ValueError
    client.subscribe(topic)  # raises ValueError
    client.on_message = on_message
    client.loop_start()
    config_debug.log(f"Successfully subcribed to topic {topic}")

    message_info = client.publish(
        "rpi", str(["STARTING"])
    )  # raises ValueError and TypeError
    if message_info.rc == 4:
        raise MQTTConnectionError

    return client


def setup_config(client):
    # Access default configurations on EpiBOX Core and save them to variables ======================

    # inform the EpiBOX App which are the current default devices
    send_default(client)
    opt = get_default()

    if not opt["channels"]:
        # if default "channels" is empty, acquire all
        channels = []
        for device in opt["devices_mac"].values():
            for i in range(1, 7):
                channels += [[device, str(i)]]
        sensors = ["-" for i in range(len(channels))]

    else:
        # default "channels" is saved in the format [[MAC1, channel1, sensor1], [MAC1, channel2, sensor2], ...]
        channels = []
        sensors = []
        try:
            for triplet in opt["channels"]:
                triplet[0] = opt["devices_mac"][
                    triplet[0]
                ]  # replace MAC ID for corresponding MAC
                channels += [triplet[:2]]
                sensors += [triplet[2]]

        except Exception as e:  # TODO what kind of errors are raised?
            config_debug.log(e)
            for tt, triplet in enumerate(opt["channels"]):
                if tt < 7:
                    triplet[0] = opt["devices_mac"]["MAC1"]
                else:
                    triplet[0] = opt["devices_mac"]["MAC2"]
                channels += [triplet[:2]]
                sensors += [triplet[2]]

    save_raw = bool(opt["save_raw"])
    service = opt["service"]

    opt["devices_mac"] = [m for m in opt["devices_mac"].values() if m != ""]

    # check if default storage is available | if not, terminate setup loop and acquisition
    opt = check_storage(client, opt)

    config_debug.log("ID: {}".format(opt["patient_id"]))
    config_debug.log("folder: {}".format(opt["initial_dir"]))
    config_debug.log("fs: {}".format(opt["fs"]))
    config_debug.log(f"save_raw: {save_raw}")
    config_debug.log(f"channels: {channels}")
    config_debug.log("devices: {}".format(opt["devices_mac"]))
    config_debug.log(f"sensors: {sensors}")
    config_debug.log(f"service: {service}")

    return opt, channels, sensors, service, save_raw


def setup_variables():

    already_notified_pause = False
    system_started = False
    t_all = []

    return t_all, already_notified_pause, system_started


def check_storage(client, opt):
    # Check if default storage is available | loop runs continuosly until it find the storage or until timeout
    # If timeout, setup loop and acquisition are terminated

    if opt["initial_dir"] == "EpiBOX Core":
        drive_path = os.path.join(Path.home(), "Documents", "EpiBOX Core")
        opt["initial_dir"] = os.path.join(drive_path, "acquisitions")

    else:
        if platform == "linux" or platform == "linux2":
            # linux
            import pwd
            drive_path = os.path.join(
                "/media", pwd.getpwuid(os.getuid())[0], opt["initial_dir"])
        elif platform == "darwin":
            # macos
            drive_path = os.path.join("/Volumes", opt["initial_dir"])
        else:
            drive_path = opt["initial_dir"]

        init_connect_time = time.time()
        last_warning = time.time()
        config_debug.log(f'Searching for storage module: {opt["initial_dir"]}')

        for i in range(1000000000):

            if not client.keepAlive:
                break

            if (time.time() - init_connect_time) > 120:
                raise StorageTimeout

            if os.path.isdir(drive_path):
                opt["initial_dir"] = os.path.join(drive_path, "acquisitions")
                break

            else:
                if time.time() - last_warning > 10:
                    message_info = client.publish(
                        "rpi", str(["INSERT STORAGE"]))
                    last_warning = time.time()
                    if message_info.rc == 4:
                        raise MQTTConnectionError

    return opt



class Error(Exception):
    """Base class for other exceptions"""

    pass


class MQTTConnectionError(Error):
    """Client is not currently connected."""

    pass


class DeviceConnectionTimeout(Error):
    """Timeout while attempting to connect to device."""

    pass


class StorageTimeout(Error):
    """Timeout while looking for storage."""

    pass


class BITalinoParameterError(Error):
    """Error raised while connecting to BITalino device due to invalid parameters."""

    pass


class ScientISSTConnectionError(Error):
    """Error raised while connecting to ScientISST device due to loss of comnmunication."""

    pass


class ScientISSTNotFound(Error):
    """ScientISST device not paired."""

    pass


class DeviceNotInAcquisitionError(Error):
    """Device not in acquisition."""

    pass


class DeviceNotIDLEError(Error):
    """Device not IDLE."""

    pass


class PlatformNotSupportedError(Error):
    """Platform not supported"""

    pass



# built-in
import ast
import os
import subprocess
import json
import shutil
from sys import platform
from pathlib import Path
from datetime import datetime

# local
from epibox import config_debug
from epibox.common.get_defaults import get_default
from epibox.exceptions.system_exceptions import (
    MQTTConnectionError,

)


def send_default(client):

    ######## Available drives ########
    listDrives = ["DRIVES"]

    if platform == "linux" or platform == "linux2":
        # linux
        import pwd
        drive_path = os.path.join("/media", pwd.getpwuid(os.getuid())[0])
        drives = [os.path.join(drive_path, d) for d in os.listdir(drive_path)]
    elif platform == "darwin":
        # macos
        drive_path = "/Volumes"
        drives = [os.path.join(drive_path, d) for d in os.listdir(drive_path)]
    else:
        import win32api
        import win32con
        import win32file
        drives = [i for i in win32api.GetLogicalDriveStrings().split('\x00') if i]
        drives = [d for d in drives if win32file.GetDriveType(
            d) == win32con.DRIVE_REMOVABLE]

    for drive in drives:
        total, _, free = shutil.disk_usage(drive)
        listDrives += ["{} ({:.1f}% livre)".format(drive,
                                                   (free / total) * 100)]

    message_info = client.publish(
        topic="rpi", qos=2, payload="{}".format(listDrives)
    )  # raises ValueError and TypeError
    if message_info.rc == 4:
        raise MQTTConnectionError

    ######## Default MAC addresses ########

    defaults = get_default()
    listMAC = defaults["devices_mac"]
    listMAC2 = json.dumps(["DEFAULT MAC"] +
                          [

        "{}".format(address) for address in list(listMAC.values())
    ]
    )

    message_info = client.publish(
        topic="rpi", qos=2, payload=listMAC2
    )  # raises ValueError and TypeError
    if message_info.rc == 4:
        raise MQTTConnectionError

    config = json.dumps(["DEFAULT CONFIG", defaults])

    message_info = client.publish(
        topic="rpi", qos=2, payload=config
    )  # raises ValueError and TypeError
    if message_info.rc == 4:
        raise MQTTConnectionError


def on_message(client, userdata, message):

    message = str(message.payload.decode("utf-8"))
    message = ast.literal_eval(message)

    if message[0] == "RESTART":
        # client.loop_stop()
        # client.keepAlive = False
        config_debug.log("Not sure what to do here yet")

    elif message[0] == "INTERRUPT":
        client.keepAlive = False

    elif message[0] == "PAUSE ACQ":
        config_debug.log("PAUSING ACQUISITION")
        client.pauseAcq = True

    elif message[0] == "RESUME ACQ":
        config_debug.log("RESUMING ACQUISITION")
        client.pauseAcq = False

    elif message[0] == "ANNOTATION":
        config_debug.log(
            "RECEIVED ANNOT {} ----------------------".format(message[1]))
        client.newAnnot = message[1] + \
            [datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')]

    elif message[0] == "TURN OFF":
        config_debug.log("TURNING OFF RPI")
        message_info = client.publish(topic="rpi", payload=str(["TURNED OFF"]))
        if message_info.rc == 4:
            raise MQTTConnectionError

    elif message[0] == "TURNED OFF":
        subprocess.run(["sudo", "shutdown", "-h", "now"])

    elif message == ["Send default"]:
        send_default(client)

    ######## New default configuration ########

    elif message[0] == "NEW CONFIG DEFAULT":
        defaults = get_default()

        for key in message[1].keys():
            defaults[key] = message[1][key]

        with open(
            os.path.join(Path.home(), "Documents",
                         "EpiBOX Core", "args.json"), "w+"
        ) as json_file:
            json.dump(defaults, json_file)

        msg = json.dumps(["RECEIVED DEFAULT"])
        message_info = client.publish(topic="rpi", qos=2, payload=msg)
        if message_info.rc == 4:
            raise MQTTConnectionError

        client.keepAlive = False



# built-in
import subprocess
import time
import sys

# local
from epibox.common.write_file import write_summary_file
from epibox import config_debug


def kill_subprocess():

    config_debug.log("  -- killing subprocess --")
    time.sleep(5)
    sys.exit()


def kill_client(client):

    try:
        client.publish("rpi", str(["STOPPED"]))
        client.loop_stop()
        config_debug.log("  -- client loop stopped --")

    except:
        pass


def close_devices(devices):

    for device in devices:
        try:
            device.close()
        except:
            device.disconnect()

    config_debug.log("  -- all devices closed --")


def stop_devices(devices):

    for device in devices:
        device.stop()

    config_debug.log("  -- all devices stopped --")


# Kill and error handling cases ===========================================================================


# | | __client__ | __devices open__ | __file open__ | __devices started__ | __acquired__|
# |:---: | :---: | :----: | :---: | :---: | :---: |
# |__kill 1__ | N | N | N | N | N |
# |__kill 2__ | Y | N | N | N | N |
# |__kill 3__ | Y | Y - | N | N | N |
# |__kill 4__ | N | Y - | N | N | N |
# |__kill 5__ | Y | Y - | Y | Y - | Y |
# |__handle 6__ | Y | Y | Y | N - | Y |
# |__kill 7__ | N | Y | Y | Y | Y |


def kill_case_1():
    # Client: N
    # Devices open: N
    # File open: N
    # Devices started: N

    kill_subprocess()


def kill_case_2(client):
    # Client: Y
    # Devices open: N
    # File open: N
    # Devices started: N

    kill_client(client)
    kill_subprocess()


def kill_case_3(client, devices):
    # Client: Y
    # Devices open: Y (might not be true for all devices)
    # File open: N
    # Devices started: N

    close_devices(devices)
    kill_client(client)
    kill_subprocess()


def kill_case_4(devices):
    # Client: N
    # Devices open: Y (might not be true for all devices)
    # File open: N
    # Devices started: N

    close_devices(devices)
    kill_subprocess()


def kill_case_5(client, devices, a_file):
    # Client: Y
    # Devices open: Y (might not be true for all devices)
    # File open: Y
    # Devices started: Y (might not be true for all devices)
    stop_devices(devices)
    close_devices(devices)
    write_summary_file(a_file.name)
    a_file.close()
    kill_client(client)
    kill_subprocess()


def handle_case_6(devices, a_file, system_started):
    # Client: Y
    # Devices open: Y
    # File open: Y
    # Devices started: N (might not be true for all devices)

    stop_devices(devices)
    write_summary_file(a_file.name)
    a_file.close()
    system_started = False

    return system_started


def kill_case_7(devices, a_file):
    # Client: N
    # Devices open: Y
    # File open: Y
    # Devices started: Y

    stop_devices(devices)
    close_devices(devices)
    write_summary_file(a_file.name)
    a_file.close()
    kill_subprocess()



# built-in
import time
from datetime import datetime
import json

# third-party

import numpy as np
import serial

# local
from epibox.devices.connect_device import connect_device
from epibox import config_debug
from epibox.exceptions.system_exceptions import (
    DeviceConnectionTimeout,
    DeviceNotIDLEError,
    DeviceNotInAcquisitionError,
    MQTTConnectionError,
    ScientISSTNotFound,
)


def start_devices(client, devices, fs, mac_channels, header):

    # Start acquisition with the biosignal acquisition devices, considering the chosen sampling ferquency
    # and channels

    dig_Out = 0
    now = datetime.now()
    sync_param = {
        "flag_sync": 0,
        "inittime": time.time(),
        "strtime": time.time(),
        "sync_time": now.strftime("%Y-%m-%d %H:%M:%S.%f").rstrip("0"),
        "dig_Out": dig_Out,
        "close_file": 0,
        "mode": 0,
        "diff": 1000,
        "save_log": 1,
        "count_a": 1000,
        "sync_append": 0,
    }
    # mode: 0 if not started acquisition yet (or if paused) and 1 otherwise (used to write in drift_log_file)

    for i in range(len(devices)):
        sync_param["sync_arr_" + chr(ord("@") + i + 1)
                   ] = np.zeros(1000, dtype=float)

    # Initialize devices
    for device in devices:

        if header["service"] == "bitalino":
            channels = [
                int(elem[1]) - 1
                for elem in mac_channels
                if elem[0] == device.address
            ]
        else:
            channels = [
                int(elem[1]) for elem in mac_channels if elem[0] == device.address
            ]

        try:
            if header["service"] == "bitalino":
                device.start(SamplingRate=fs, analogChannels=channels)

            elif header["service"] == "scientisst":
                device.start(sample_rate=fs, channels=channels)

        except Exception as e:
            config_debug.log(e)
            raise DeviceNotIDLEError

    now = datetime.now()

    config_debug.log("start {now}")
    sync_param["sync_time"] = now.strftime("%Y-%m-%d %H:%M:%S.%f")

    message_info = client.publish("rpi", str(["ACQUISITION ON"]))
    if message_info.rc == 4:  # MQTT_ERR_NO_CONN
        raise MQTTConnectionError

    return sync_param


def connect_devices(
    client, devices, opt, already_timed_out,
):

    # This script attempts to connect to the default biosignal acquisition devices in a continuous loop.
    # The loop stops only if:
    #       - connection is successful
    #       - timeout is achieved (2min)

    for mac in opt["devices_mac"]:

        init_connect_time = time.time()
        config_debug.log(f"Searching for Module... {mac}")

        for i in range(10000000):

            if not client.keepAlive:
                break

            if (time.time() - init_connect_time) > 120:
                raise DeviceConnectionTimeout

            try:
                connected = False
                connected, devices = connect_device(
                    mac, client, devices, opt["service"]
                )

                if not (connected and mac in [d.address for d in devices]):
                    if time.time() - init_connect_time > 3 * i:
                        timeout_json = json.dumps(
                            ["TRYING TO CONNECT", "{}".format(mac)]
                        )
                        message_info = client.publish("rpi", timeout_json)
                        if message_info.rc == 4:
                            raise MQTTConnectionError

                else:
                    break

            except serial.SerialException as e:
                time.sleep(2)
                config_debug.log(f"Serial connection refused: {e}")
                if not already_timed_out and (time.time() - init_connect_time > 3 * i):
                    timeout_json = json.dumps(["TIMEOUT", "{}".format(mac)])
                    message_info = client.publish("rpi", timeout_json)
                    if message_info.rc == 4:
                        raise MQTTConnectionError
                    already_timed_out = True

                continue

            except ScientISSTNotFound as e:
                time.sleep(2)
                config_debug.log(f"Connection refused: {e}")
                if not already_timed_out and (time.time() - init_connect_time > 3 * i):
                    timeout_json = json.dumps(["TIMEOUT", "{}".format(mac)])
                    message_info = client.publish("rpi", timeout_json)
                    if message_info.rc == 4:
                        raise MQTTConnectionError
                    already_timed_out = True

                continue

            except Exception as e:
                time.sleep(2)
                config_debug.log(f"Connection refused: {e}")
                if not already_timed_out and (time.time() - init_connect_time > 3 * i):
                    timeout_json = json.dumps(["TIMEOUT", "{}".format(mac)])
                    message_info = client.publish("rpi", timeout_json)
                    if message_info.rc == 4:
                        raise MQTTConnectionError
                    already_timed_out = True

                continue

    return devices


def pause_devices(client, devices):

    try:
        for device in devices:
            device.stop()
    except Exception as e:
        config_debug.log(e)
        raise DeviceNotInAcquisitionError

    message_info = client.publish("rpi", str(["PAUSED"]))
    if message_info.rc == 4:
        raise MQTTConnectionError

    return devices



# built-in
from copy import copy

# third-party
import numpy as np

from epibox import config_debug
from epibox.exceptions.system_exceptions import DeviceNotInAcquisitionError


def read_modules(devices, mac_channels, sensors, header):

    t = np.array([])
    t_display = np.array([])
    channels2get = [int(mac_chn[1]) - 1 for mac_chn in mac_channels]

    for i, device in enumerate(devices):

        try:
            if header["service"] == "bitalino":
                t_read = device.read(100)
            elif header["service"] == "scientisst":
                t_read = device.read(convert=False, matrix=True)

        except Exception as e:
            config_debug.log(e)
            raise DeviceNotInAcquisitionError

        t_read = np.delete(
            t_read, np.arange(1, 5), axis=1
        )  # remove digital channels
        t_nseq = t_read[:, 1:]  # remove nSeq column (add again in the end)

        # else:
        #     seq = []
        #     t_nseq = []

        #     for f in t_read:

        #         frame = f.toMap()
        #         seq += [frame["sequence"]]

        #         if len(t_nseq) == 0:
        #             t_nseq_aux = np.reshape(np.array(frame["analog"]), (1, -1))
        #             t_nseq = np.reshape(
        #                 np.take(t_nseq_aux, channels2get), (1, -1))
        #         else:
        #             t_nseq_aux = np.reshape(
        #                 np.take(
        #                     np.reshape(
        #                         np.array(frame["analog"]), (1, -1)), channels2get
        #                 ),
        #                 (1, -1),
        #             )
        #             t_nseq = np.vstack((t_nseq, t_nseq_aux))

        # t_read = np.hstack((np.reshape(np.array(seq), (-1, 1)), t_nseq))

        # config_debug.log(t_nseq)

        n = 0
        display_aux = np.array([])

        for j, chn in enumerate(mac_channels):

            if chn[0] == device.address:

                if header["save_raw"]:
                    signal_type = "RAW"

                else:
                    signal_type = sensors[j]

                r = header["resolution"][device.address][n + 1]
                t_aux = get_transform(
                    t_nseq[:, n], signal_type, r
                )  # receives and returns a column with 100 samples

                n += 1
                if len(display_aux) == 0:
                    display_aux = copy(t_aux)

                else:
                    display_aux = np.concatenate((display_aux, t_aux), axis=1)

        if len(t_display) == 0:
            t_display = copy(display_aux)
            t = np.concatenate(
                (np.reshape(t_read[:, 0], (-1, 1)), display_aux), axis=1)

        else:
            t_display = np.concatenate((t_display, display_aux), axis=1)
            aux = np.concatenate(
                (np.reshape(t_read[:, 0], (-1, 1)), display_aux), axis=1
            )
            t = np.concatenate((t, aux), axis=1)

    t_str = "\n".join(" ".join("%0.1f" % x for x in y) for y in t)
    return t, t_str, t_display


def get_transform(raw, signal_type, res):

    if signal_type == "ECG":
        aux = list(
            map(lambda x: (((x / (2**res) - 0.5) * 3.3) / 1100) * 1000, raw))

    elif signal_type == "EEG":
        aux = list(
            map(lambda x: (((x / (2**res) - 0.5) * 3.3) / 41782) * (10 ** (6)), raw)
        )

    elif signal_type == "EOG":
        aux = list(
            map(lambda x: (((x / (2**res) - 0.5) * 3.3) / 2040) * 1000, raw))

    elif signal_type == "EMG":
        aux = list(
            map(lambda x: (((x / (2**res) - 0.5) * 3.3) / 1009) * 1000, raw))

    elif signal_type == "PZT":
        aux = list(map(lambda x: (x / (2**res) - 0.5) * 100, raw))

    elif signal_type == "EDA":
        aux = list(map(lambda x: (((x / (2**res)) * 3.3) / 0.132), raw))

    else:
        aux = list(raw)

    return np.reshape(np.asarray(aux), (-1, 1))



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()


test_requirements = []

setup(
    author="Ana Sofia Carmo",
    author_email="anascacais@gmail.com",
    python_requires=">=3.6, <3.10",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="EpiBOX is a Raspberry Pi tool for easy signal acquisition.",
    entry_points={
        "console_scripts": [
            "epibox=epibox.cli:main",
        ],
    },
    install_requires=[
        "bitalino==1.2.1",
        "numpy==1.21.0; sys_platform != 'win32'",
        "numpy; sys_platform == 'win32'",
        "pywin32; sys_platform == 'win32'",
        "paho-mqtt==1.5.1",
        "pyserial==3.5",
        "pybluez2; sys_platform != 'darwin'",
        "scientisst-sense==1.0.5",
        "scipy",
        "bump"
    ],
    license="MIT license",
    long_description_content_type="text/markdown",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords=["epibox", "signal acquisition", "Raspberry Pi"],
    name="epibox",
    packages=find_packages(include=["epibox", "epibox.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/anascacais/epibox",
    version="3.0.0",
    zip_safe=False,
)



# built-in
import os

# local
from epibox import config_debug


def create_folder(initial_dir, nb):

    # Create folder with patient ID unless it already exists
    directory = os.path.join(initial_dir, nb)

    if not os.path.exists(directory):
        os.makedirs(directory)
        config_debug.log(f"Created patient directory -- {directory}")

    else:
        config_debug.log(f"Directory -- {directory} -- already exists")

    return directory



from epibox import config_debug


def get_header(
    filename,
    file_time,
    file_date,
    devices,
    mac_channels,
    sensors,
    fs,
    save_raw,
    service,
):

    header_dict = {}
    resolution = {}
    columns = []
    fmt = []  # get format to save values in txt

    for n_device, device in enumerate(devices):

        fmt += ["%i"]
        header_dict[device.address] = {}
        header_dict[device.address]["sensor"] = []
        for i, elem in enumerate(mac_channels):
            if elem[0] == device.address:
                header_dict[device.address]["sensor"] += [sensors[i]]
                if save_raw:
                    fmt += ["%i"]
                else:
                    fmt += ["%.2f"]
        header_dict[device.address]["device name"] = "Device " + \
            str(n_device + 1)

        aux = [elem[1]
               for elem in mac_channels if elem[0] == device.address]
        header_dict[device.address]["column"] = ["nSeq"] + aux
        columns += header_dict[device.address]["column"]

        header_dict[device.address]["start time"] = file_time
        header_dict[device.address]["device connection"] = device.address

        header_dict[device.address]["channels"] = [
            int(elem[1]) for elem in mac_channels if elem[0] == device.address
        ]

        header_dict[device.address]["date"] = file_date
        # header_dict[device.address]["firmware version"] = device.version()

        if service == "bitalino":
            header_dict[device.address]["device"] = "bitalino_rev"
            aux = [10, 10, 10, 10, 6, 6]  # resolution
        elif service == "scientisst":
            header_dict[device.address]["device"] = "scientisst_sense"
            aux = [12, 12, 12, 12, 12, 12]
        else:
            raise ValueError("Device not recognized")

        aux2 = [1 for elem in mac_channels if elem[0] == device.address]
        header_dict[device.address]["resolution"] = [4] + [
            aux[i] for i in range(len(aux2))
        ]
        resolution[device.address] = header_dict[device.address]["resolution"]

        header_dict[device.address]["sampling rate"] = fs

        if save_raw:
            header_dict[device.address]["label"] = [
                "RAW" for elem in mac_channels if elem[0] == device.address
            ]
        else:
            header_dict[device.address]["label"] = [
                sensors[i]
                for i, elem in enumerate(mac_channels)
                if elem[0] == device.address
            ]

    header = {"resolution": resolution,
              "save_raw": save_raw, "service": service}

    config_debug.log(f"# {header_dict} \n")
    config_debug.log(f"# {columns} \n")

    filename.write("# " + str(header_dict) + "\n")
    filename.write("\t".join(columns) + "\n")

    return tuple(fmt), header
